#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <glib.h>
#include <unistd.h>



#include "common.h"

static int
internal_tcp_set_buffer(const tcp_con_t *con,int type,int max,int min,int *actual){
  int rc;
  int size;

  if ( (!con) || (!actual) )
    return -EINVAL;
  if ( (type != SO_RCVBUF) && (type != SO_SNDBUF) )
    return -EINVAL;

  size=max;
  rc=setsockopt(con->soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
  if (rc<0) {
    size=min;
    rc=setsockopt(con->soc, SOL_SOCKET, type, (void*)&size, sizeof(int));
    if (rc<0) {
      dbg_out("Can not set buffer size:%s(%d)\n",strerror(errno),errno);
      *actual=size;
      return -errno;
    }
  }
  *actual=size;
  return 0;
}
static int
tcp_set_buffer(const tcp_con_t *con){
  int rc;
  int size=0;

  if (!con)
    return -EINVAL;

  rc=internal_tcp_set_buffer(con,SO_RCVBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set recv buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }
  size=0;
  rc=internal_tcp_set_buffer(con,SO_SNDBUF,_MSG_BUF_SIZE,_MSG_BUF_MIN_SIZE,&size);
  if (size < _MSG_BUF_MIN_SIZE) {
    err_out("Can not set send buf size:%s (%d)\n",
	    strerror(-rc),
	    -rc);
    return rc;
  }

  return 0;
}
static int
parse_request(request_msg_t *req,const char *request_string){
  size_t len;
  long int_val;
  char *sp=NULL;
  char *ep=NULL;
  char *buffer;
  ssize_t remains;
  int rc=0;

  if  ( (!request_string) || (!req) )
    return -EINVAL;

  buffer=strdup(request_string);
  if (!buffer)
    return -ENOMEM;
  len=strlen(request_string);
  _assert(len>0);
  remains=len;
  /*
   * パケット番号
   */
  sp=buffer;
  ep=memchr(sp, ':', remains);
  if (!ep) {
    rc=-EINVAL;
    goto err_out;
  }

  *ep='\0';
  remains =len - ((unsigned long)ep-(unsigned long)buffer);
  if (remains<=0) {
    rc=-EINVAL;
    goto err_out;
  }
  ++ep;
  int_val=strtol(sp, (char **)NULL, 16);
  req->pkt_no=int_val;
  dbg_out("pktno:%ld\n",req->pkt_no);
  sp=ep;

  /*
   * ファイルID
   */
  ep=memchr(sp, ':', remains);
  if (!ep) {
    rc=-EINVAL;
    goto err_out;
  }
  *ep='\0';
  remains =len - ((unsigned long)ep-(unsigned long)buffer);
  if (remains<=0) {
    rc=-EINVAL;
    goto err_out;
  }
  ++ep;
  int_val=strtol(sp, (char **)NULL, 16);
  req->fileid=int_val;
  dbg_out("fileid:%d(%x)\n",req->fileid,req->fileid);
  sp=ep;

  /*
   * オフセット
   */
  int_val=strtol(sp, (char **)NULL, 16);
  req->offset=int_val;
  dbg_out("offset:%d(%x)\n",req->offset,req->offset);
err_out:
  free(buffer);
  return rc;
}
static int
tcp_transfer_file(tcp_con_t *con,const char *path,const size_t size,off_t offset){
  int fd;
  int rc;
  char buff[TCP_FILE_BUFSIZ];
  ssize_t read_len;
  ssize_t file_remains;
  ssize_t soc_remains;
  size_t write_len;
  size_t total_write;
  char *wp;

  if ( (!con) || (!path) )
    return -EINVAL;

  fd=open(path,O_RDONLY);
  if (fd<0)
    return -errno;

  rc=lseek(fd, offset, SEEK_SET);
  if (rc<0) {
    rc=-errno;
    goto close_out;
  }
  total_write=0;
  file_remains=size;

  while(file_remains>0) {
    read_len=read(fd,buff,TCP_FILE_BUFSIZ);
    if (read_len<0) {
      rc=-errno;
      goto close_out;
    }
    file_remains -= read_len;
    soc_remains=read_len;
    wp=buff;
    while(soc_remains>0) {
      write_len=send(con->soc,wp,soc_remains,MSG_NOSIGNAL);
      if (write_len<0) {
	if (errno==EINTR)
	  continue;
	goto close_out;
      }
      wp += write_len;
      total_write += write_len;
      soc_remains -= write_len;
    }
    dbg_out("transfer %s %d/%d(%d)\n",path,total_write,size,file_remains);
  }
  rc=0;
 close_out:
  close(fd);
  if (rc<0)
    dbg_out("Can not send file:%s %d\n",strerror(errno),errno);
  return rc;
}
int
tcp_setup_client(const char *ipaddr,int port,int *ret_soc){
  int sock=-1;
  struct sockaddr_in server;
  struct hostent *host;

  if ( (!ipaddr) || (!ret_soc) )
    return -EINVAL;

  sock=socket(AF_INET, SOCK_STREAM, 0);

  if (sock<0) {
    err_out("Can not create socket: %s(errno:%d)\n",strerror(errno),errno);
    return -errno;
  }

 server.sin_family = AF_INET;
 server.sin_port = htons(port);
 server.sin_addr.s_addr = inet_addr(ipaddr);

 if (server.sin_addr.s_addr == 0xffffffff) {
   /* ホスト名で指定してきた */
   host = gethostbyname(ipaddr);
   if (!host) {
     err_out("Can not resolve addr:%s %s(errno:%d)\n",ipaddr,strerror(errno),errno);
     return -errno;
   }
   memcpy((void *)&server.sin_addr,host->h_addr,host->h_length);
 }

 if (connect(sock, (struct sockaddr *)&server, sizeof(server))<0){
   err_out("Can not connect addr:%s port:%d %s (errno:%d)\n",ipaddr,port,strerror(errno),errno);
   return -errno;
 }
   
 *ret_soc=sock; 
 return 0;
}
int
tcp_set_ipmsg_bufsiz(int soc) {
 tcp_con_t con;

 con.soc=soc;
 return tcp_set_buffer(&con);
}
gpointer
ipmsg_tcp_recv_thread(gpointer data){
  ssize_t recv_len;
  size_t addr_len;
  fd_set rd_set;
  fd_set wr_set;
  int rc;
  struct timeval tmout_val;
  tcp_con_t *con;
  char recv_buf[_MSG_BUF_SIZE];
  int count=200;

  con=(tcp_con_t *)data;
  if (!con) {
    err_out("No connection recived\n");
    g_assert(con);
    return NULL;
  }

  rc=tcp_set_buffer(con);
  if (rc<0)
    return NULL;

  while(1) {
  retry:
    FD_ZERO(&rd_set);
    FD_ZERO(&wr_set);
    FD_SET(con->soc,&rd_set);
    FD_SET(con->soc,&wr_set);
    dbg_out("select\n");
    tmout_val.tv_sec=TCP_SELECT_SEC;
    tmout_val.tv_usec=0;
    rc=select(con->soc+1,&rd_set,NULL,NULL,&tmout_val);
    if (rc>0)
      dbg_out("selectOK\n");
    if (rc<0) {
      if (errno == EINTR)
	goto retry;
      return NULL;
    }
    
    if (FD_ISSET(con->soc,&rd_set)) {
      memset(recv_buf, 0, sizeof(recv_buf));
      addr_len=sizeof(con->client);
      recv_len=recvfrom(con->soc,recv_buf,_MSG_BUF_SIZE,MSG_PEEK,(struct sockaddr *)&(con->client),&addr_len);

      if (recv_len<=0) {
	if (errno)
	  err_out("Can not peek message %s(errno:%d)\n",strerror(errno),errno);
	close(con->soc);
	free(con);
	return NULL;
      }
      dbg_out("tcp peek read:%d %s\n",recv_len,recv_buf);
      addr_len=sizeof(con->client);
      recv_len=recvfrom(con->soc,recv_buf,recv_len,0,(struct sockaddr *)&(con->client),&addr_len);
      dbg_out("tcp read:%d %s\n",recv_len,recv_buf);
      if (recv_len>0) {
	msg_data_t msg;
	request_msg_t req;
	char *path;
	size_t size;

	recv_buf[recv_len-1]='\0';
	memset(&req,0,sizeof(request_msg_t));
	init_message_data(&msg);
	parse_message(&msg,recv_buf,recv_len);
	parse_request(&req,msg.message);

	if (refer_attach_file(req.pkt_no,req.fileid,(const char **)&path,&size)<0) {
	  err_out("Can not find message:pktno %ld id : %d\n",
		  req.pkt_no,
		  req.fileid);
	}else{
	  dbg_out("transfer:%s (%d)\n",path,size);
	  tcp_transfer_file(con,path,size,0);
	}
	release_message_data(&msg);
	break;
      }
    }
    --count;
    if (!count)
      break;
  }
  close(con->soc);
  free(con);

  return NULL;
}
gpointer 
ipmsg_tcp_server_thread(gpointer data){
  size_t len;
  int sock;
  int tcp_socket;
  struct sockaddr_in addr;
  int port;
  int reuse;

  tcp_socket = socket(AF_INET, SOCK_STREAM, 0);
  if (tcp_socket < 0) {
    err_out("Can not create socket:%s (%d)\n",strerror(errno),errno);
    return NULL;
  }
  port=hostinfo_refer_ipmsg_port();
  memset(&addr,0,sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);

  reuse=1;
  if (setsockopt(tcp_socket,SOL_SOCKET, SO_REUSEADDR, (const char *)&reuse, sizeof(reuse))<0) {
    err_out("Can not set sock opt:%s (%d)\n",strerror(errno),errno);
    return NULL;
  }

  if (bind(tcp_socket, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
    err_out("Can not bind socket:%s (%d)\n",strerror(errno),errno);
    exit(errno);
  }

  if (listen(tcp_socket, TCP_LISTEN_LEN) != 0) {
    err_out("Can not listen socket:%s (%d)\n",strerror(errno),errno);
    return NULL;
  }

  while (1) {
    tcp_con_t *client_con;
    GThread *handler;

    client_con=malloc(sizeof(tcp_con_t));
    g_assert(client_con);
    dbg_out("accept\n");
    len = sizeof(client_con->client);
    sock = accept(tcp_socket, (struct sockaddr *)&(client_con->client), &len);
    if (sock < 0) {
      err_out("Can not accept:%s (%d)\n",strerror(errno),errno);
      break;
    }
    client_con->soc=sock;
    handler=g_thread_create(ipmsg_tcp_recv_thread,
                            client_con,
			    FALSE,
			    NULL);
    if (!handler) {
      err_out("Can not create handler thread.\n");
      close(sock);
    }
  } 
  return NULL;
}
